#include <fstream.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include <dos.h>
#include <new.h>
#include "sin.h"
#include "cos.h"

//#define A  0x1000000L   // room's width
//#define B  0x2000000L   // room's length
//#define C  0x0800000L   // room's height
#define TEXTURE_SIZE  16384

long x, y;              // texture coordinates
char far *texture;      // active texture;
char far *texture1;     // floor
char far *texture2;     // wall1
char far *texture3;     // wall2
char far *texture4;     // sky
char pal[1024];

long A = 0x1000000L;   // room's width
long B = 0x2000000L;   // room's length
long C = 0x0800000L;   // room's height

long x_c = 0, y_c = -B + 0x20000L, z_c = -C + 0x400000L;
int count = 0, angle = 25;
int curbank = 0;

struct vesainfo {
  unsigned int  modeattr        ; // mode attributes
  unsigned char winaattr        ; // window A attributes
  unsigned char winbattr        ; // window B attributes
  unsigned int  granularity     ; // window granularity in K
  unsigned int  windowsize      ; // window size in K
  unsigned int  winaseg         ; // window A start segment
  unsigned int  winbseg         ; // window B start segment
  unsigned long farwindow       ; // far window pointer
  unsigned int  bytesperscan    ; // bytes per scan line
  unsigned int  scrwidth        ; // screen width in pixels
  unsigned int  scrheight       ; // screen height in pixels
  unsigned char charwidth       ; // char width in pixels
  unsigned char charheight      ; // char height in pixels
  unsigned char Nummemplanes    ; // number of memory planes
  unsigned char bitsperpixel    ; // bits per pixel
  unsigned char numofbanks      ; // number of banks
  unsigned char memmodel        ; // memory model
  unsigned char banksize        ; // size of bank in K
  unsigned char numimgpages     ; // number of image pages
  unsigned char reserved[255]   ;
} info;

void Intersect(long x_c, long y_c, long z_c,
               long l, long m, long n)
{
  long y1, z1, x2, z2, x3, y3, r;

  // 1 - X-intersection

  if (l)
  {
    if (l > 0)
      r = (A - x_c) / l;
    else
      r = (A + x_c) / -l;
    y1 = y_c + r * m;
    z1 = z_c + r * n;
  }

  // 2 - Y-intersection

  if (m)
  {
    if (m > 0)
      r = (B - y_c) / m;
    else
      r = (B + y_c) / -m;
    x2 = x_c + r * l;
    z2 = z_c + r * n;
  }

  // 3 - Z-intersection

  if (n)
  {
    if (n > 0)
      r = (C - z_c) / n;
    else
      r = (C + z_c) / -n;
    x3 = x_c + r * l;
    y3 = y_c + r * m;
  }

  if ((y1 <= B) && (y1 >= -B) && (z1 <= C) && (z1 >= -C))
  {
    x = y1;
    y = z1;
    texture = texture2;
  }
  else
  if ((x2 <= A) && (x2 >= -A) && (z2 <= C) && (z2 >= -C))
  {
    x = x2;
    y = z2;
    texture = texture3;
  }
  else
  if ((x3 <= A) && (x3 >= -A) && (y3 <= B) && (y3 >= -B))
  {
    x = x3;
    y = y3;
    if (n > 0)
      texture = texture4;
    else
      texture = texture1;
  }
  else
    texture = texture4;

  x >>= 16;
  y >>= 16;

  if (texture == texture4)
  {
    x += count;
    y += count;
  }

  if (x < 0) x += 128;
  if (y < 0) y += 128;

  x &= 127;
  y &= 127;
}

int LoadTexture(char *filename, char far *&texturename)
{
  ifstream in(filename, ios::binary);

  if (!in) return 1;

  texturename = new char far[TEXTURE_SIZE];
  in.seekg(1078, ios::beg);
  in.read((char *)texturename, TEXTURE_SIZE);
  in.close();

  return 0;
}

int LoadPalette(char *filename)
{
  ifstream in(filename, ios::binary);

  if (!in) return 1;

  in.seekg(54, ios::beg);
  in.read(pal, 1024);
  in.close();

  return 0;
}

void SetPalette()
{
  for (int i = 1; i < 256; i++)
  {
    outportb(0x3C8, i);
    outportb(0x3C9, pal[i * 4 + 2] / 4);
    outportb(0x3C9, pal[i * 4 + 1] / 4);
    outportb(0x3C9, pal[i * 4] / 4);
  }
}

void SetMode(int mode)
{
  struct REGPACK reg;

  reg.r_ax = mode;
  intr(0x10, &reg);
}

void Range()
{
  if (x_c < -A + 0x200000L) x_c = -A + 0x200000L;
  if (y_c < -B + 0x200000L) y_c = -B + 0x200000L;
  if (z_c < -C + 0x100000L) z_c = -C + 0x100000L;

  if (x_c > A - 0x200000L) x_c = A - 0x200000L;
  if (y_c > B - 0x200000L) y_c = B - 0x200000L;
  if (z_c > C - 0x100000L) z_c = C - 0x100000L;
}

void my_handler()
{
  cout << "Error: not enough memory!\n";
  exit(1);
}

unsigned SetVesaMode(unsigned mode)
{
  unsigned status;

  _ES = FP_SEG((void far *)&info);
  _DI = FP_OFF((void far *)&info);

  asm
  {
    mov ax, 4F01h
    mov cx, mode
    int 10h
    mov ax, 4F02h
    mov bx, mode
    int 10h
    mov status, ax
  }

  return status;
}

void PutPixel(int x, int y, char r, char g, char b)
{
  int bytes=info.bytesperscan;
  unsigned int color1;

  asm
  {
    mov ah, g
    mov al, b
    mov color1, ax
    push 0A000h
    pop es
    xor dx, dx
    mov ax, x
    mov bx, 3
    mul bx
    mov bx, ax
    mov ax, y
    mul bytes
    add ax, bx
    jnc @1
    inc dx
@1: cmp dx, curbank
    je @2

    push ax
    mov curbank, dx
    mov ax, 4F05h
    xor bx, bx
    int 10h
    pop ax

@2: mov di, ax
    mov ax, color1
    stosw
    mov al, r
    stosb
  }
}

void main()
{
  int i = 0, j, l, l2, m2;
  unsigned char c, r, g, b, color;
  set_new_handler(my_handler);

  LoadPalette("059-013.bmp");
  LoadTexture("222-001.bmp", texture1);
  LoadTexture("067-013.bmp", texture2);
  LoadTexture("252-002.bmp", texture3);
  LoadTexture("sky.bmp", texture4);

//  SetMode(0x13);
  if (SetVesaMode(0x212)!=0x004F)
    cout << "Error initializing mode\n";
//  SetPalette();

  do
  {
    asm  in al, 60h
    asm  mov c, al

    switch (c)
    {
      case 77: if (!angle) angle = 99; else angle--; break;
      case 75: if (angle == 99) angle = 0; else angle++; break;
      case 72: x_c += 16 * cos_table[angle];
               y_c += 16 * sin_table[angle]; break;
      case 80: x_c -= 16 * cos_table[angle];
               y_c -= 16 * sin_table[angle]; break;
      case 32: z_c += 0x100000L; break;
      case 46: z_c -= 0x100000L; break;
    }

    Range();

    for (j = 0; j < 480; j++)
      for (i = 0; i < 640; i++)
      {
        l = 320 - i;

        l2 = int((277 * cos_table[angle] - l * sin_table[angle]) >> 16);
        m2 = int((277 * sin_table[angle] + l * cos_table[angle]) >> 16);

        Intersect(x_c, y_c, z_c, l2, m2, 240 - j);

        color = *(texture + (unsigned)y * 128 + (unsigned)x);
        r = pal[color * 4 + 2];
        g = pal[color * 4 + 1];
        b = pal[color * 4];

        PutPixel(i, j, r, g, b);

/*        asm  push 0A000h
        asm  pop es
        asm  mov di, j
        asm  mov ax, di
        asm  shl di, 8
        asm  shl ax, 6
        asm  add di, ax
        asm  add di, i
        asm  mov al, color
        asm  stosb        */
      }

    while (kbhit()) getch();

    count += 4;
    count &= 127;

  } while (c != 1);

  SetMode(0x03);
}
